/*
 * Copyright 2021 Sony Corporation
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <debug_impl.h>
#include <linux/types.h>
#include <include/sony_fs.h>
#include <vsprintf.h>

#include "vbo_frame.h"
#include "vbo_frame_info.h"
#include "data_variation_resolver_client.h"

#define READ_BLOCK 4
#define PKT_PATH_LENGTH 128
#define FRAMEBUFFER_FIRSTLINE_WIDTH 256  // 64dot * 4
#define FRAMEBUFFER_WIDTH 15360          // 3840dot * 4
#define FRAME_HEADER_CRC_WIDTH 2
#define TID_WORD_LENGTH 4
#define DATA_NAME_LENGTH 16
#define PARTITION_NAME_LENGTH 16
#define JSON_PATH_LENGTH 128

#define HEADER_SIZE_IDX 7
#define FRAME_SIZE_IDX_1 8
#define FRAME_SIZE_IDX_2 9
#define FRAME_SIZE_IDX_3 10
#define FRAME_SIZE_IDX_4 11
#define AMOUNT_IN_FRAME_IDX 27
#define CRC_FRAME0_IDX_1 30
#define CRC_FRAME0_IDX_2 31
#define CRC_IDX_1 126
#define CRC_IDX_2 127

#define DATA_PATH "/pq/be_data/"
#define JSON_NAME "/data_variation.json"
#define FOLDER_NAME "/files/"
#define INDEX_NAME "index.txt"
#define FDAT_PERSIST_PATH "/sony/display/fdat/"
#define FDAT_CONFIG_PATH "/display/fdat/"
#define FDAT_PKT_DEFAULT "fdat_be_1.pkt"
#define FDAT_NAME "FDAT"
#define PERSIST_PART_NAME "persist"
#define SONY_CONFIG_PART_NAME "sony_config"

#define ERROR_PATH "ERROR"
#define NO_DATA_PATH ""

static void get_be_frm_path(uint32_t frame_id, uint8_t packetcount, struct ST_FRAME_HEADER *frame_header, uint32_t pkt_path_size, char *pkt_path, uint32_t partition_name_size, char *partition_name);
static void get_fdat_path(uint32_t pkt_path_size, char *partition_name, char *fdat_path, char *pkt_path);
static void get_data_path(char *tid_small, char *pkt_path);
static int set_data_to_frame(char *buf, uint32_t *frame_index, int32_t pkt_size, uint32_t buffer_size, uint32_t frame_id, char *data);
static int set_header_and_footer(uint8_t packet_sum, uint32_t frame_index, uint32_t header_size, uint32_t frame_size, uint32_t frame_id, uint32_t buffer_size, struct ST_FRAME_HEADER *frame_header, char *data);
static int output_debug_log(unsigned long long *frame_buffer, uint32_t frame_id);

static void get_be_frm_path(uint32_t frame_id, uint8_t packetcount, struct ST_FRAME_HEADER *frame_header, uint32_t pkt_path_size, char *pkt_path, uint32_t partition_name_size, char *partition_name)
{
	char tid[TID_WORD_LENGTH] = {0};
	char tid_small[TID_WORD_LENGTH] = {0};
	int32_t pkt_size = -1;

	if (pkt_path == NULL) {
		UBOOT_ERROR("pkt_path NULL\n");
	} else if (frame_header == NULL) {
		UBOOT_ERROR("frame_header NULL\n");
		strncpy(pkt_path, ERROR_PATH, pkt_path_size - 1);
	} else if (frame_id >= FRAME_HEADER_NUM_MAX) {
		UBOOT_ERROR("frame_id Error\n");
		strncpy(pkt_path, ERROR_PATH, pkt_path_size - 1);
	} else if (packetcount >= TID_NUM_SIZE_MAX) {
		UBOOT_ERROR("packetcount Error\n");
		strncpy(pkt_path, ERROR_PATH, pkt_path_size - 1);
	} else {
		memset(pkt_path, 0, pkt_path_size);

		//Change TID from uppercase to lowercase
		for (uint32_t j = 0; j < TID_WORD_LENGTH; j++) {
			tid[j] = frame_header->packets_frame[frame_id].tid_frame[packetcount][j];
			if (('A' <= tid[j]) && (tid[j] <= 'Z'))
				tid_small[j] = tid[j] + ('a' - 'A');
			else
				tid_small[j] = tid[j];
		}

		if (strncmp(tid, FDAT_NAME, TID_WORD_LENGTH) == 0) {
			//Get PKT Path(persist)
			strncpy(partition_name, PERSIST_PART_NAME, partition_name_size - 1);
			get_fdat_path(pkt_path_size, partition_name, FDAT_PERSIST_PATH, pkt_path);
			if (strcmp(pkt_path, NO_DATA_PATH) != 0)
				pkt_size = ext4_f_size_mmc(partition_name, pkt_path);
			if ((strcmp(pkt_path, NO_DATA_PATH) == 0) || (pkt_size == -1)) {
				//Get PKT Path(config)
				memset(pkt_path, 0, pkt_path_size);
				strncpy(partition_name, SONY_CONFIG_PART_NAME, partition_name_size - 1);
				get_fdat_path(pkt_path_size, partition_name, FDAT_CONFIG_PATH, pkt_path);
			}
			UBOOT_INFO("fdat frm :%s\n", pkt_path);
		} else {
			//JSON read
			get_data_path(tid_small, pkt_path);
			strncpy(partition_name, DATA_PART_NAME, partition_name_size - 1);
		}
	}
}

static void get_fdat_path(uint32_t pkt_path_size, char *partition_name, char *fdat_path, char *pkt_path)
{
	char *fdat_name = NULL;
	char index_txt_path[PKT_PATH_LENGTH] = {0};
	int32_t load_ret = -1;
	int32_t fdat_index_size = -1;
	int32_t snprintf_len = 0;

	if (pkt_path == NULL) {
		UBOOT_ERROR("pkt_path NULL\n");
		return;
	}
	if (fdat_path == NULL) {
		UBOOT_ERROR("fdat_path NULL\n");
		strncpy(pkt_path, NO_DATA_PATH, pkt_path_size - 1);
		return;
	}
	if (partition_name == NULL) {
		UBOOT_ERROR("partition_name NULL\n");
		strncpy(pkt_path, NO_DATA_PATH, pkt_path_size - 1);
		return;
	}

	//index.txt Read
	snprintf_len = snprintf(index_txt_path, PKT_PATH_LENGTH, "%s%s", fdat_path, INDEX_NAME);
	if ((snprintf_len < 0) || (snprintf_len >= PKT_PATH_LENGTH)) {
		UBOOT_ERROR("snprintf fail fdat index path\n");
		strncpy(pkt_path, NO_DATA_PATH, pkt_path_size - 1);
		return;
	}
	fdat_index_size = ext4_f_size_mmc(partition_name, index_txt_path);
	if (fdat_index_size == -1) {
		UBOOT_ERROR("FileSize Error fdat_index_size : %s\n", index_txt_path);
		//Get PKT Path
		snprintf_len = snprintf(pkt_path, PKT_PATH_LENGTH, "%s%s", fdat_path, FDAT_PKT_DEFAULT);
		if ((snprintf_len < 0) || (snprintf_len >= PKT_PATH_LENGTH)) {
			UBOOT_ERROR("snprintf fail fdat default path\n");
			strncpy(pkt_path, NO_DATA_PATH, pkt_path_size - 1);
			return;
		}
	} else {
		fdat_name = (char *)malloc(fdat_index_size);
		if (fdat_name == NULL) {
			UBOOT_ERROR("fdat_name malloc error\n");
			strncpy(pkt_path, NO_DATA_PATH, pkt_path_size - 1);
			return;
		}
		memset((void *)fdat_name, 0, fdat_index_size);
		load_ret = ext4_f_load_mmc(partition_name, index_txt_path, fdat_name);
		if (load_ret == -1) {
			UBOOT_ERROR("FileOpen_Error fdat_name : %s\n", index_txt_path);
			strncpy(pkt_path, NO_DATA_PATH, pkt_path_size - 1);
			free(fdat_name);
			return;
		}
		//Get PKT Path
		snprintf_len = snprintf(pkt_path, PKT_PATH_LENGTH, "%s%s", fdat_path, fdat_name);
		if ((snprintf_len < 0) || (snprintf_len >= PKT_PATH_LENGTH)) {
			UBOOT_ERROR("snprintf fail fdat path\n");
			strncpy(pkt_path, NO_DATA_PATH, pkt_path_size - 1);
			free(fdat_name);
			return;
		}
		free(fdat_name);
	}
}

static void get_data_path(char *tid_small, char *pkt_path)
{
	char json_path[JSON_PATH_LENGTH] = {0};
	char data_name[DATA_NAME_LENGTH] = {0};
	char *json_string = NULL;
	int32_t json_size = -1;
	int32_t load_ret = -1;
	int32_t snprintf_len = 0;

	if (pkt_path == NULL) {
		UBOOT_ERROR("pkt_path NULL\n");
		return;
	}
	if (tid_small == NULL) {
		UBOOT_ERROR("tid_small NULL\n");
		return;
	}

	//JSON read
	snprintf_len = snprintf(json_path, PKT_PATH_LENGTH, "%s%s%s", DATA_PATH, tid_small, JSON_NAME);
	if ((snprintf_len < 0) || (snprintf_len >= PKT_PATH_LENGTH)) {
		UBOOT_ERROR("snprintf fail json path\n");
		return;
	}

	json_size = ext4_f_size_mmc(DATA_PART_NAME, json_path);
	if (json_size == -1) {
		UBOOT_ERROR("FileSize Error json_path : %s\n", json_path);
		return;
	}

	json_string = (char *)malloc(json_size);
	if (json_string == NULL) {
		UBOOT_ERROR("json_string malloc error\n");
		return;
	}
	memset((void *)json_string, 0, json_size);
	load_ret = ext4_f_load_mmc(DATA_PART_NAME, json_path, json_string);
	if (load_ret == -1) {
		UBOOT_ERROR("FileOpen_Error json_path : %s\n", json_path);
		free(json_string);
		return;
	}

	//Get Dataname
	find_data(json_string, DATA_NAME_LENGTH, data_name);
	if (strnlen(data_name, DATA_NAME_LENGTH) == 0) {
		UBOOT_ERROR("Get Dataname Error : %s\n", tid_small);
		free(json_string);
		return;
	}
	//Get PKT Path
	snprintf_len = snprintf(pkt_path, PKT_PATH_LENGTH, "%s%s%s%s", DATA_PATH, tid_small, FOLDER_NAME, data_name);
	if ((snprintf_len < 0) || (snprintf_len >= PKT_PATH_LENGTH)) {
		UBOOT_ERROR("snprintf fail pkt path\n");
		free(json_string);
		return;
	}
	free(json_string);
}


int get_beboot_vb1_frame_num(void)
{
	return load_frame_info_num();
}

int draw_beboot_vb1_frame(uint32_t frame_id, unsigned long long *frame_buffer, uint32_t buffer_size)
{
	struct ST_FRAME_HEADER frame_header;
	char *data;
	char *buf;
	char pkt_path[PKT_PATH_LENGTH];
	char partition_name[PARTITION_NAME_LENGTH] = {0};
	uint8_t error_num;
	uint8_t packetcount = 0;
	uint8_t packet_sum = 0;
	uint8_t packetcount_max = 0;
	uint32_t frame_index;
	uint32_t header_size;
	uint32_t frame_size = 0;
	int32_t pkt_size = -1;
	int32_t load_ret = -1;

	if (frame_buffer == NULL) {
		UBOOT_ERROR("frame_buffer NULL\n");
		return RESULT_ERROR;
	}

	data = (char *)malloc(buffer_size);
	if (data == NULL) {
		UBOOT_ERROR("data malloc error\n");
		return RESULT_ERROR;
	}
	memset((void *)data, 0, buffer_size);

	// get header
	frame_header.num = frame_id + 1;
	error_num = make_vb1_frame_header(&frame_header);
	if (error_num != 0) {
		UBOOT_ERROR("make_vb1_frame_header error : %d\n", error_num);
		free(data);
		return RESULT_ERROR;
	}

	// set packets
	packetcount_max = frame_header.param[frame_id][AMOUNT_IN_FRAME_IDX];
	header_size = frame_header.param[frame_id][HEADER_SIZE_IDX];

	if (frame_id == 0)
		frame_index = FRAMEBUFFER_WIDTH + header_size;
	else
		frame_index = header_size;

	while (packetcount < packetcount_max) {
		get_be_frm_path(frame_id, packetcount, &frame_header, sizeof(pkt_path), &pkt_path[0], sizeof(partition_name), &partition_name[0]);
		//ERROR case
		if (strcmp(pkt_path, ERROR_PATH) == 0) {
			free(data);
			return RESULT_ERROR;
		}
		if (strcmp(pkt_path, NO_DATA_PATH) == 0) {
			//Data not found
			packetcount++;
		} else {
			UBOOT_DEBUG("target frm :%s\n", pkt_path);
			pkt_size = ext4_f_size_mmc(partition_name, pkt_path);
			if (pkt_size == -1) {
				UBOOT_ERROR("pkt file can not open :%s\n", pkt_path);
				packetcount++;
			} else {
				buf = (char *)malloc(pkt_size);
				if (buf == NULL) {
					UBOOT_ERROR("buf malloc error\n");
					free(data);
					return RESULT_ERROR;
				}
				memset((void *)buf, 0, pkt_size);

				load_ret = ext4_f_load_mmc(partition_name, pkt_path, buf);
				if ((strncmp(pkt_path, NO_DATA_PATH, PKT_PATH_LENGTH) != 0) && (load_ret == -1)) {
					UBOOT_ERROR("packetfile open error\n");
					free(buf);
					free(data);
					return RESULT_ERROR;
				}
				frame_size = frame_size + pkt_size;
				error_num = set_data_to_frame(buf, &frame_index, pkt_size, buffer_size, frame_id, data);
				if (error_num != RESULT_NORMAL) {
					free(data);
					free(buf);
					return RESULT_ERROR;
				}
				packetcount++;
				packet_sum++;
				free(buf);
			}
		}
	}

	error_num = set_header_and_footer(packet_sum, frame_index, header_size, frame_size, frame_id, buffer_size, &frame_header, data);
	if (error_num != RESULT_NORMAL) {
		free(data);
		return RESULT_ERROR;
	}
	memcpy((void *)frame_buffer, (void *)data, (size_t)buffer_size);

	error_num = output_debug_log(frame_buffer, frame_id);
	if (error_num != RESULT_NORMAL) {
		free(data);
		return RESULT_ERROR;
	}
	free(data);
	return RESULT_NORMAL;
}

static int set_data_to_frame(char *buf, uint32_t *frame_index, int32_t pkt_size, uint32_t buffer_size, uint32_t frame_id, char *data)
{
	char read[READ_BLOCK];
	uint8_t count = 0;
	uint32_t data_index = 0;

	if (buf == NULL) {
		UBOOT_ERROR("buf NULL");
		return RESULT_ERROR;
	}
	if (frame_index == NULL) {
		UBOOT_ERROR("frame_index NULL");
		return RESULT_ERROR;
	}

	while (pkt_size > 0) {
		if (pkt_size < 4) {
			memcpy(read, &buf[data_index], pkt_size);
			count = pkt_size;
			pkt_size = pkt_size - 4;
		} else {
			memcpy(read, &buf[data_index], 4);
			count = 4;
			pkt_size = pkt_size - 4;
			data_index = data_index + 4;
		}
		if ((*frame_index + READ_BLOCK) > buffer_size) {
			UBOOT_ERROR("frm size is over than buffer size packet");
			return RESULT_ERROR;
		}
		// GBRA -> BGRA
		switch (count) {
		case 1:
			UBOOT_DEBUG("count %d\n", count);
			data[*frame_index]   = 0x00;
			data[*frame_index+1] = read[0];
			data[*frame_index+2] = 0x00;
			data[*frame_index+3] = 0x00;
			break;
		case 2:
			UBOOT_DEBUG("count %d\n", count);
			data[*frame_index]   = read[1];
			data[*frame_index+1] = read[0];
			data[*frame_index+2] = 0x00;
			data[*frame_index+3] = 0x00;
			break;
		case 3:
			UBOOT_DEBUG("count %d\n", count);
			data[*frame_index]   = read[1];
			data[*frame_index+1] = read[0];
			data[*frame_index+2] = read[2];
			data[*frame_index+3] = 0x00;
			break;
		case 4:
			data[*frame_index]   = read[1];
			data[*frame_index+1] = read[0];
			data[*frame_index+2] = read[2];
			data[*frame_index+3] = read[3];
			break;
		}
		*frame_index += READ_BLOCK;
		if ((frame_id == 0) && ((*frame_index % FRAMEBUFFER_WIDTH) == FRAMEBUFFER_FIRSTLINE_WIDTH))
			*frame_index = *frame_index + FRAMEBUFFER_WIDTH - FRAMEBUFFER_FIRSTLINE_WIDTH;
	}
	return RESULT_NORMAL;
}

static int set_header_and_footer(uint8_t packet_sum, uint32_t frame_index, uint32_t header_size, uint32_t frame_size, uint32_t frame_id, uint32_t buffer_size, struct ST_FRAME_HEADER *frame_header, char *data)
{
	uint16_t crc;
	uint16_t crc_size;
	uint32_t n = 0;
	uint32_t data_index = frame_index;

	if (frame_header == NULL) {
		UBOOT_ERROR("frame_header NULL");
		return RESULT_ERROR;
	}
	if (data == NULL) {
		UBOOT_ERROR("data NULL");
		return RESULT_ERROR;
	}

	// set footer
	for (int j = 0; j < 3; j++) {
		data_index += READ_BLOCK;
		if (frame_id == 0) {
			if ((data_index % FRAMEBUFFER_WIDTH) == FRAMEBUFFER_FIRSTLINE_WIDTH)
				data_index = data_index + FRAMEBUFFER_WIDTH - FRAMEBUFFER_FIRSTLINE_WIDTH;
		}
	}
	data[data_index]   = frame_header->ikey[1];
	data[data_index+1] = frame_header->ikey[0];
	data[data_index+2] = frame_header->ikey[2];
	data[data_index+3] = frame_header->ikey[3];

	// set Frame Size
	frame_header->param[frame_id][FRAME_SIZE_IDX_1] = (frame_size >> 24) & (0xFF);
	frame_header->param[frame_id][FRAME_SIZE_IDX_2] = (frame_size >> 16) & (0xFF);
	frame_header->param[frame_id][FRAME_SIZE_IDX_3] = (frame_size >> 8)  & (0xFF);
	frame_header->param[frame_id][FRAME_SIZE_IDX_4] = (frame_size)       & (0xFF);

	// set Amount in Frame
	frame_header->param[frame_id][AMOUNT_IN_FRAME_IDX] = packet_sum;

	// set HeaderCRC
	crc_size = header_size - FRAME_HEADER_CRC_WIDTH;
	crc = calc_vb1_header_crc(&frame_header->param[frame_id][0], crc_size);

	if (frame_id == 0) {
		data_index = FRAMEBUFFER_WIDTH;
		frame_header->param[frame_id][CRC_FRAME0_IDX_1] = (crc >> 8) & (0xFF);
		frame_header->param[frame_id][CRC_FRAME0_IDX_2] = (crc)      & (0xFF);
	} else {
		data_index = 0;
		frame_header->param[frame_id][CRC_IDX_1] = (crc >> 8) & (0xFF);
		frame_header->param[frame_id][CRC_IDX_2] = (crc)      & (0xFF);
	}

	// set header
	while (n < frame_header->param[frame_id][HEADER_SIZE_IDX]) {
		if ((data_index + READ_BLOCK) > buffer_size) {
			UBOOT_ERROR("frm size is over than buffer size header");
			return RESULT_ERROR;
		}
		data[data_index]   = frame_header->param[frame_id][n+1];
		data[data_index+1] = frame_header->param[frame_id][n];
		data[data_index+2] = frame_header->param[frame_id][n+2];
		data[data_index+3] = frame_header->param[frame_id][n+3];

		data_index += READ_BLOCK;
		n += READ_BLOCK;
		if (frame_id == 0) {
			if ((data_index % FRAMEBUFFER_WIDTH) == FRAMEBUFFER_FIRSTLINE_WIDTH)
				data_index = data_index + FRAMEBUFFER_WIDTH - FRAMEBUFFER_FIRSTLINE_WIDTH;
		}
	}
	return RESULT_NORMAL;
}

static int output_debug_log(unsigned long long *frame_buffer, uint32_t frame_id)
{
	if (frame_buffer == NULL) {
		UBOOT_ERROR("frame_buffer NULL\n");
		return RESULT_ERROR;
	}

	// Debug Log
	if (frame_id == 0) {
		for (int32_t a = 0; a <= 2; a += 1)
			UBOOT_DEBUG("frame_buffer [%09d]:%016llx\n", a*8, frame_buffer[a]);
		for (int32_t a = 1919; a <= 1925; a += 1)
			UBOOT_DEBUG("frame_buffer [%09d]:%016llx\n", a*8, frame_buffer[a]);
	} else {
		for (int32_t a = 0; a <= 20; a += 1)
			UBOOT_DEBUG("frame_buffer [%09d]:%016llx\n", a*8, frame_buffer[a]);
	}

	return RESULT_NORMAL;
}
