/*
 * Copyright 2022 Sony Corporation
 */
#include "sony_panel.h"
#include <asm/gpio.h>
#include <linux/delay.h>
#include <time.h>

#include "autodet.h"

#include <common.h>
#include <utility.h>
#include <part.h>
#include <cli.h>
#include <partition.h>
#include <command.h>
#include "iniutility.h"
#include "qhb_internal.h"
#include "system_impl.h"
#include <dm.h>
#include <i2c.h>

/*
 * This function enable panel power and read panel ID.
 * This function needs GPIO overlay and product config,
 * and panel initialize function dto.
 */
enum paneldet_state {
	DETSTATE_INIT = 0,
	DETSTATE_PANELDET,
	DETSTATE_1ST_POWER,	// For Panel power(LCD) or 12V power(OLED)
	DETSTATE_READPID,
	DETSTATE_2ND_POWER,	// For Backlight power(LCD) or 24V power(OLED)
	DETSTATE_BEDET,

	DETSTATE_FINISH,
	DETSTATE_INVALID,
};

static struct panel_det_priv priv;

int paneldet_getPaneldetResult(void)
{
	return priv.isPanelLess;
}

int paneldet_getPanelId(char *buf)
{
	int ret;

	if (priv.panelId != NULL) {
		memcpy(buf, priv.panelId, sizeof(priv.panelId));
		ret = 0;
	} else {
		printf("%s[%u]:: Panel id is null!!\n", __func__, __LINE__);
		ret = -ENODEV;
	}

	return ret;
}

//0:FE Only 1:With 4KBE 2:With 8KBE
int paneldet_getBEType(void)
{
	int ret;

	if (priv.isBE != NULL) {
		if (strncmp(priv.isBE, ISBE_NONE, sizeof(ISBE_NONE)) == 0)
			ret = BEDET_NONE;
		else if (strncmp(priv.isBE, ISBE_4K, sizeof(ISBE_4K)) == 0)
			ret = BEDET_4K;
		else if (strncmp(priv.isBE, ISBE_8K, sizeof(ISBE_8K)) == 0)
			ret = BEDET_8K;
		else
			ret = BEDET_INVALID;
	} else {
		printf("%s[%u]:: isBE is null!!\n", __func__, __LINE__);
		ret = BEDET_INVALID;
	}

	return ret;
}

static int panel_id_read(void)
{
	int ret;
	ofnode node;
	const char *buf;

	ret = select_sony_panel_overlay(priv.panelI2Cbus.bus, priv.panelI2Cbus.addr, priv.panelI2Cbus.reg);

	if (ret == 0) {
		node = ofnode_by_compatible(ofnode_null(), "sony,paneldet");
		if (!ofnode_valid(node)) {
			printf("PANEL_DET: Fail to get panel det dts!!!!!\n");
			return AUTODET_NO_RULES;
		}
		buf = ofnode_read_string(node, "panel_id");
		memset(priv.panelId, 0x00, sizeof(priv.panelId));
		strncpy(priv.panelId, buf, (sizeof(priv.panelId)-1));
	}

	return ret;
}

static int is_be_read(void)
{
	int ret = 0;
	ofnode node;
	const char *buf;

	node = ofnode_by_compatible(ofnode_null(), "sony,paneldet");
	if (!ofnode_valid(node)) {
		printf("PANEL_DET: Fail to get panel det dts!!!!!\n");
		return -1;
	}
	buf = ofnode_read_string(node, "isBE");
	memset(priv.isBE, 0x00, sizeof(priv.isBE));
	strncpy(priv.isBE, buf, (sizeof(priv.isBE)-1));

	return ret;
}

int paneldet_setPanelPower(bool enable)
{
	if (enable == true) {
		mdelay(priv.panelOnPreDelay);
		gpio_direction_output(priv.panelPowerIoNum[0], priv.panelPowerIoNum[1]);
		mdelay(priv.panelOnPostDelay);
	} else {
		mdelay(priv.panelOffPreDelay);
		gpio_direction_output(priv.panelPowerIoNum[0], !(priv.panelPowerIoNum[1]));
		mdelay(priv.panelOffPostDelay);
	}

	return 0;
}

int paneldet_setBacklightPower(bool enable)
{
	if (enable == true) {
		mdelay(priv.blOnPreDelay);
		gpio_direction_output(priv.blPowerIoNum[0], priv.blPowerIoNum[1]);
		mdelay(priv.blOnPostDelay);
	} else {
		mdelay(priv.blOffPreDelay);
		gpio_direction_output(priv.blPowerIoNum[0], !(priv.blPowerIoNum[1]));
		mdelay(priv.blOffPostDelay);
	}

	return 0;
}

static void set_pid_error(int error_val)
{
	if (priv.do_panelId_check == true) {
		if (error_val == AUTODET_NO_RULES)
			add_bootargs("PID_Result", "PID_Result=1", false);
		else if (error_val == AUTODET_NO_DATA)
			add_bootargs("PID_Result", "PID_Result=2", false);
		else
			add_bootargs("PID_Result", "PID_Result=0", false);
	} else {
		add_bootargs("PID_Result", "PID_Result=0", false);
	}
}

static int paneldet_findmode(void)
{
	enum paneldet_state state = DETSTATE_INIT;
	bool isDone = false;
	bool temp_isPanelLess = true;
	unsigned int ioVal;
	int ret = 0;
	int pid_ret = 0;

	while (!isDone) {
		switch (state) {
		case DETSTATE_INIT:
			if (priv.do_panelDet_check)
				state = DETSTATE_PANELDET;
			else
				state = DETSTATE_1ST_POWER;
			break;
		case DETSTATE_PANELDET:
			gpio_direction_input(priv.panelDetIoNum[0]);
			ioVal = gpio_get_value(priv.panelDetIoNum[0]);
			if (ioVal == priv.panelDetIoNum[1]) {
				temp_isPanelLess = false;
				state = DETSTATE_1ST_POWER;
			} else {
				temp_isPanelLess = true;
				set_pid_error(0);
				state = DETSTATE_BEDET;
			}
			break;
		case DETSTATE_1ST_POWER:
			ret = paneldet_setPanelPower(true);
			if (ret != 0) {
				printf("##### Fail to enable panel/12V power\n");
				temp_isPanelLess = true;
				set_pid_error(0);
				state = DETSTATE_BEDET;
			} else if (true == priv.do_panelId_read) {
				state = DETSTATE_READPID;
			} else {
				temp_isPanelLess = true;
				set_pid_error(0);
				state = DETSTATE_BEDET;
			}
			break;
		case DETSTATE_READPID:
			// call_autodet
			pid_ret = panel_id_read();
			if (pid_ret == AUTODET_NO_RULES) {
				printf("##### Undefined PID\n");
				temp_isPanelLess = true;
			} else if (pid_ret == AUTODET_NO_DATA) {
				printf("##### Fail to read PID\n");
				temp_isPanelLess = true;
			} else {
				temp_isPanelLess = false;
			}
			set_pid_error(pid_ret);
			state = DETSTATE_2ND_POWER;
			break;
		case DETSTATE_2ND_POWER:
			ret = paneldet_setBacklightPower(true);
			if (ret != 0) {
				printf("##### Fail to enable backlight/24V power\n");
				temp_isPanelLess = true;
			}
			state = DETSTATE_BEDET;
			break;
		case DETSTATE_BEDET:
			select_sony_adc_overlay(0x04);
			state = DETSTATE_FINISH;
			break;
		case DETSTATE_FINISH:
			priv.isPanelLess = temp_isPanelLess;
			if (true == priv.isPanelLess)
				strncpy(priv.panelId, "FFFFFFFFFFFFFFFF", sizeof(priv.panelId));

			isDone = true;
			printf("##### result: isPanelLess=[%d]\n", priv.isPanelLess);
			ret = is_be_read();
			if (ret != 0)
				printf("##### BE_DET ERROR\n");

			break;
		default:
			printf("[PANEL_DET][WARN] Unexpected det state. Assume panel less mode.\n");
			priv.isPanelLess = true;
			isDone = true;
			break;
		}
	}

	return 0;
}

static int parse_devicetree(struct udevice *dev)
{
	ofnode node;
	int ret;
	uint32_t val;
	uint32_t temparray[4];

	node = ofnode_by_compatible(ofnode_null(), "sony,paneldet");
	if (!ofnode_valid(node)) {
		printf("PANEL_DET: Fail to get panel det dts!!!!!\n");
		return -1;
	}

	// Operation properties
	priv.do_panelDet_check = false;
	ret = ofnode_read_u32(node, "doPanelDet", &val);
	if (ret != 0)
		printf("PANEL_DET: Fail to get [doPanelDet]!!!!!\n");
	else if (val == 1)
		priv.do_panelDet_check = true;

	priv.do_panelId_read = false;
	ret = ofnode_read_u32(node, "doPanelIdRead", &val);
	if (ret != 0)
		printf("PANEL_DET: Fail to get [doPanelIdRead]!!!!!\n");
	else if (val == 1)
		priv.do_panelId_read = true;

	priv.do_panelId_check = false;
	ret = ofnode_read_u32(node, "checkPanelIdError", &val);
	if (ret != 0)
		printf("PANEL_DET: Fail to get [checkPanelIdError]!!!!!\n");
	else if (val == 1)
		priv.do_panelId_check = true;

	if (true == priv.do_panelDet_check) {
		ret = ofnode_read_u32_array(node, "panelDetIoNum", priv.panelDetIoNum, 2);
		if (ret != 0)
			printf("PANEL_DET: Fail to get [panelDetIoNum]!!!!!\n");
		else
			gpio_request(priv.panelDetIoNum[0], "PANEL_DET");
	}

	if (true == priv.do_panelId_read) {
		ret = ofnode_read_u32_array(node, "panelI2Cbus", temparray, 4);
		if (ret != 0) {
			printf("PANEL_DET: Fail to get [panelI2Cbus]!!!!!\n");
		} else {
			priv.panelI2Cbus.bus = temparray[0];
			priv.panelI2Cbus.addr = temparray[1];
			priv.panelI2Cbus.reg = temparray[2];
			priv.panelI2Cbus.reg_size = temparray[3];
		}
	}

	// Panel power IOs
	ret = ofnode_read_u32_array(node, "panelPowerIoNum", priv.panelPowerIoNum, 2);
	if (ret != 0)
		printf("PANEL_DET: Fail to get [panelPowerIoNum]!!!!!\n");
	else
		gpio_request(priv.panelPowerIoNum[0], "PANEL_PWR");

	ret = ofnode_read_u32(node, "panelOnPreDelay", &priv.panelOnPreDelay);
	if (ret != 0)
		printf("PANEL_DET: Fail to get [panelOnPreDelay]!!!!!\n");

	ret = ofnode_read_u32(node, "panelOnPostDelay", &priv.panelOnPostDelay);
	if (ret != 0)
		printf("PANEL_DET: Fail to get [panelOnPostDelay]!!!!!\n");

	ret = ofnode_read_u32(node, "panelOffPreDelay", &priv.panelOffPreDelay);
	if (ret != 0)
		printf("PANEL_DET: Fail to get [panelOffPreDelay]!!!!!\n");

	ret = ofnode_read_u32(node, "panelOffPostDelay", &priv.panelOffPostDelay);
	if (ret != 0)
		printf("PANEL_DET: Fail to get [panelOffPostDelay]!!!!!\n");

	// Backlight power IOs
	ret = ofnode_read_u32_array(node, "blPowerIoNum", priv.blPowerIoNum, 2);
	if (ret != 0)
		printf("PANEL_DET: Fail to get [blPowerIoNum]!!!!!\n");
	else
		gpio_request(priv.blPowerIoNum[0], "BL_ON");

	ret = ofnode_read_u32(node, "blOnPreDelay", &priv.blOnPreDelay);
	if (ret != 0)
		printf("PANEL_DET: Fail to get [blOnPreDelay]!!!!!\n");

	ret = ofnode_read_u32(node, "blOnPostDelay", &priv.blOnPostDelay);
	if (ret != 0)
		printf("PANEL_DET: Fail to get [blOnPostDelay]!!!!!\n");

	ret = ofnode_read_u32(node, "blOffPreDelay", &priv.blOffPreDelay);
	if (ret != 0)
		printf("PANEL_DET: Fail to get [blOffPreDelay]!!!!!\n");

	ret = ofnode_read_u32(node, "blOffPostDelay", &priv.blOffPostDelay);
	if (ret != 0)
		printf("PANEL_DET: Fail to get [blOffPostDelay]!!!!!\n");

	return 0;
}

static int sony_panel_det_probe(struct udevice *dev)
{
	printf("[Panel det] Probed!!!!\n");
	return 0;
}

static const struct udevice_id panel_det_ids[] = {
	{ .compatible = "sony,paneldet" },
	{ }
};

U_BOOT_DRIVER(sony_panel_det) = {
	.name = "sony_paneldet",
	.id = UCLASS_PANEL,
	.of_match = panel_det_ids,
//	.ops = &panel_det_dummy_ops,
	.ofdata_to_platdata = parse_devicetree,
	.probe = sony_panel_det_probe,
	.priv_auto_alloc_size = sizeof(struct panel_det_priv),
};

void disable_panel_test_pattern(void)
{
	struct udevice *dev = NULL;
	int bus = 0x02;		//I2C bus: PANEL_I2C (M6L I2C ch2)
	int slave = 0x30;	//TCON SlvAddr: 0x60 -> 7bit(0x30)
	int addr = 0x0012;	//SubAddr(2Byte): 0x0012
	int len = 2;		//2byte
	unsigned char wbuf[2] = { 0x00, 0x01 };	//2ByteF 0x0001 [1:0]
	unsigned char rbuf[2] = { 0x00, 0x00 };	//2Byte
	unsigned int cnt = 0;

	//Not Panel Less and Not Self Run
	if ((priv.isPanelLess == false) && !is_panel_self_run()) {

		while (cnt < 5) {

			//i2c setup
			if (i2c_get_chip_for_busnum(bus, slave, len, &dev) == 0) {
				//i2c Write
				if (dm_i2c_write(dev, addr, wbuf, len) != 0) {
					printf("[Panel det] Fail Disable panel test pattern : i2c err\n");
					return;
				}
				//i2c Read
				if (dm_i2c_read(dev, addr, rbuf, len) == 0) {
					if ((rbuf[0] == 0x00) && (rbuf[1] == 0x01)) {
						printf("[Panel det] Disable panel test pattern\n");
						break;
					}
				}
			}
			mdelay(1);
			cnt++;
		}
		if (cnt >= 5)
			printf("[Panel det] Fail Disable panel test pattern : Retry limit\n");
	}
}

int paneldet_callback(void)
{
	struct udevice dummydev; // Temporary

	parse_devicetree(&dummydev);

	priv.isPanelLess = true;
	paneldet_findmode();

	disable_panel_test_pattern();

	return 0;
}
