/*
 * Copyright 2022 Sony Corporation
 */

#include <common.h>
#include <command.h>
#include <dm.h>
#include <dm/of_access.h>
#include <asm/gpio.h>
#include <adc.h>
#include <mtk_dtbo.h>
#include <mtk-pm.h>
#include <boot_impl.h>
#include <powerkey_fdr_cmd.h>
#include <sony_gpio_util.h>

int do_check_powerkey_fdr(cmd_tbl_t *cmdtp, int flag, int argc, char *const argv[])
{
	struct udevice *dev;
	unsigned int data;
	int ret, cnt = 0, dev_cnt = 0;

	// FDR usecase is only from AC on
	if (pm_get_boot_reason() != PM_BR_AC) {
		printf("%s Skip powerkey check\n", __func__);
		return CMD_RET_SUCCESS;
	}
#if defined(ADC_PORT_POWERKEY_DEV) /* use a defined device in a header file */

	ret = uclass_get_device_by_name(UCLASS_ADC, ADC_PORT_POWERKEY_DEV, &dev);
	if (ret) {
		printf("ERR : %s, Unknown ADC device %s\n", __func__, ADC_PORT_POWERKEY_DEV);
		return CMD_RET_FAILURE;
	}
	dev_cnt = 1;

#else /* use auto detected device from device tree */

	/* find available adc devices */
	ret = uclass_first_device_err(UCLASS_ADC, &dev);
	if (ret) {
		printf("ERR : %s, No available ADC device\n", __func__);
		return CMD_RET_FAILURE;
	}
	/* check multiple adc device */
	do {
		dev_cnt++;
		ret = uclass_next_device(&dev);
		if (ret)
			return CMD_RET_FAILURE;
	} while (dev);

	if (dev_cnt > 1) {
		printf("ERR : %s, Has multiple ADC device\n", __func__);
		return CMD_RET_FAILURE;
	}
#endif

	ret = uclass_first_device_err(UCLASS_ADC, &dev);

	printf("Use device:%s port:%d for %s\n", dev->name, ADC_PORT_POWERKEY_NUM, __func__);

	ret = adc_channel_single_shot(dev->name, ADC_PORT_POWERKEY_NUM, &data);

	/* get adc value to check pressed and holded powerkey */

	if (data < ADC_PORT_POWERKEY_PRESSED_LEVEL_LOW) {
		printf("Powerkey is pressed\n");
	} else {
		printf("Powerkey is not pressed\n");
		return CMD_RET_SUCCESS;
	}

	/* get adc value to check holded powerkey */

	do {
		mdelay(1000);
		cnt++;
		ret = adc_channel_single_shot(dev->name, ADC_PORT_POWERKEY_NUM, &data);
	} while (cnt <= POWERKEY_FDR_CHECK_PERIOD_SEC && data < ADC_PORT_POWERKEY_PRESSED_LEVEL_LOW);

	if (cnt >= POWERKEY_FDR_CHECK_PERIOD_SEC) {

#if defined(POWERKEY_FDR_USE_RECOVERY)
		// LED
		int ret;
		uint32_t gpio_led;

		printf("FDR LED\n");
		ret = find_gpio_line_name("LED_W", &gpio_led);
		if (ret != 0) {
			printf("Not Found WLED ret=%d\n", ret);
		} else {
			printf("Found LED ret=%d gpio_led=%u\n", ret, gpio_led);
			gpio_request(gpio_led, "FDR");
			gpio_direction_output(gpio_led, 1);
		}

		/* executing run_command("wipeuserdata", 0) */
		if (wipe_user_data() != 0) {
			printf("Write misc data to erase userdata failed!\n");
			return CMD_RET_FAILURE;
		}
		printf("Write misc data to erase userdata success\n");

		mdelay(8000); /* waiting for end of sync & indicate LED to user several seconds */
		printf("Finish to write and reset\n");
		do_reset(NULL, 0, 0, NULL);
#else
		/* executing run_command("wipeuserdataboot", 0) */
		if (wipe_user_data_in_uboot() != 0) {
			printf("Erase userdata in uboot failed!\n");
			return CMD_RET_FAILURE;
		}
		printf("Erase userdata in uboot success\n");
#endif

	}

	return CMD_RET_SUCCESS;
}

U_BOOT_CMD(check_pkey_fdr, 1, 1, do_check_powerkey_fdr,
	"Check Powerkey hold and Run FDR\n",
	"Check Powerkey hold and Run FDR when Powerkey is pressed and holded for a while\n"
);


